--author: igor29381
--v1.1
--date 14.10.17

StorageUnit = {};

local StorageUnit_mt = Class(StorageUnit);

function StorageUnit:new(id, parent)
	local self = {};
	setmetatable(self, StorageUnit_mt);
	self.rootNode = id;
	self.parent = parent;
	local tipTriggerNode = getChild(id, "tipTrigger");
	local siloTriggerNode = getChild(id, "siloTrigger");
	if tipTriggerNode > 0 and siloTriggerNode > 0 then
		self.fillUnit = {level=0, currentFillType=0};
		--tipTrigger
		local tipTrigger = TipTrigger:new(g_server ~= nil, g_client ~= nil);
		tipTrigger.initPricingDynamics = function() end;
		tipTrigger.className = "TipTrigger";
		tipTrigger:load(tipTriggerNode);
		tipTrigger:register(true);
		tipTrigger.acceptedFillTypes = parent.acceptedFillTypes;
		tipTrigger.fillTypePrices = parent.fillTypePrices;
		tipTrigger.fillTypePriceInfo = parent.fillTypePriceInfo;
		local toolTypes = {true, true, true};
		tipTrigger.allowedToolTypes = {};
		for fillType,_ in pairs(parent.acceptedFillTypes) do
			tipTrigger.allowedToolTypes[fillType] = toolTypes;
		end;
		tipTrigger.addFillLevelFromTool = function(trigger, object, fillDelta, fillType, toolType)
			if fillDelta > 0 then
				if parent.acceptedFillTypes[fillType] then
					fillDelta = math.min(fillDelta, self.fillUnit.capacity - self.fillUnit.level);
					self:updateFillLevel(fillDelta, fillType);
				end;
				return fillDelta;
			end;
		end;
		tipTrigger.getNotAllowedText = function(trigger, fillable, toolType)
			if not trigger.suppressWarnings then
				local fillTypes = fillable:getCurrentFillTypes();
				if fillTypes ~= nil and next(fillTypes) ~= nil then
					local hasGenerallyAcceptedFillType = false;
					local acceptedFillType;
					local notAcceptedFillTypes = "";
					local notAcceptedToolFillTypes = "";
					for _,fillType in pairs(fillTypes) do
						if trigger.acceptedFillTypes[fillType] then
							hasGenerallyAcceptedFillType = true;
							if trigger.allowedToolTypes[fillType][toolType] then
								if acceptedFillType == nil then
									acceptedFillType = fillType;
								end;
							else
								if notAcceptedToolFillTypes ~= "" then
									notAcceptedToolFillTypes = notAcceptedToolFillTypes .. ", ";
								end;
								notAcceptedToolFillTypes = notAcceptedToolFillTypes .. FillUtil.fillTypeIndexToDesc[fillType].nameI18N;
							end;
						else
							if notAcceptedFillTypes ~= "" then
								notAcceptedFillTypes = notAcceptedFillTypes .. ", ";
							end;
							notAcceptedFillTypes = notAcceptedFillTypes .. FillUtil.fillTypeIndexToDesc[fillType].nameI18N;
						end;
					end;
					if acceptedFillType == nil then
						if hasGenerallyAcceptedFillType then
							return string.format(g_i18n:getText("warning_notAcceptedTool"), notAcceptedToolFillTypes);
						else
							return string.format(g_i18n:getText("notAcceptedHere"), notAcceptedFillTypes);
						end;
					else
						if self.fillUnit.capacity <= self.fillUnit.level then
							return string.format(g_i18n:getText("warning_noMoreFreeCapacity"), FillUtil.fillTypeIndexToDesc[acceptedFillType].nameI18N);
						end;
						if self.fillUnit.level > 1 and acceptedFillType ~= self.fillUnit.currentFillType then
							return string.format(g_i18n:getText("notAcceptedHere"), FillUtil.fillTypeIndexToDesc[acceptedFillType].nameI18N);
						end;
					end;
				end;
			end;
		end;
		tipTrigger.getAllowFillTypeFromTool = function(trigger, fillType, toolType)
			if parent.acceptedFillTypes[fillType] and (self.fillUnit.level < 1 or fillType==self.fillUnit.currentFillType or self.fillUnit.currentFillType==FillUtil.FILLTYPE_UNKNOWN)
			and self.fillUnit.capacity > self.fillUnit.level and trigger.allowedToolTypes[fillType][toolType] then
				return true;
			end;
			return false;
		end;
		tipTrigger.getTipInfoForTrailer = function(trigger, trailer, tipReferencePointIndex)
			local minDistance, bestPoint = trigger:getTipDistanceFromTrailer(trailer, tipReferencePointIndex);
			local isAllowed = false;
			local fillTypes = trailer:getCurrentFillTypes();
			if fillTypes then
				for _,fillType in pairs(fillTypes) do
					if trigger:getAllowFillTypeFromTool(fillType, TipTrigger.TOOL_TYPE_TRAILER) then
						isAllowed = true;
						break;
					end;
				end;
			end;
			return isAllowed, minDistance, bestPoint;
		end;
		self.tipTrigger = tipTrigger;
		--siloTrigger
		local siloTrigger = SiloTrigger:new(g_server ~= nil, g_client ~= nil);
		siloTrigger.dropParticleSystems = {};
		siloTrigger:load(siloTriggerNode);
		siloTrigger.dropParticleSystems = nil;
		siloTrigger:register(true);
		siloTrigger.fillTypes = {};
		for fillType,_ in pairs(parent.acceptedFillTypes) do
			siloTrigger.fillTypes[fillType] = fillType;
		end;
		siloTrigger.getFillLevel = function(trigger)
			return self.fillUnit.level;
		end;
		siloTrigger.getIsActivatable = function(trigger)
			if not trigger.isEnabled then
				return false;
			end;
			if trigger.isFilling then
				if trigger.siloTrailer:getRootAttacherVehicle() == g_currentMission.controlledVehicle then
					return true;
				end;
			else
				if trigger.siloTrailer ~= nil and trigger.activeTriggers >= 4 then
					local trailer = trigger.siloTrailer;
					if trailer:getRootAttacherVehicle() ~= g_currentMission.controlledVehicle then
						return false;
					end;
					local currentFillType = trailer.fillUnits[1].currentFillType;
					local fillLevel = trailer:getFillLevel();
					if self.fillUnit.level > 0 and (trigger.selectedFillType==currentFillType or currentFillType==FillUtil.FILLTYPE_UNKNOWN) and fillLevel < trailer:getCapacity() then
						return true;
					end;
				end;
			end;
			return false;
		end;
		siloTrigger.onActivateObject = function(trigger)
			if not trigger.isFilling then
				trigger:setIsFilling(true, trigger.selectedFillType);
			else
				trigger:setIsFilling(false, FillUtil.FILLTYPE_UNKNOWN);
			end;
			g_currentMission:addActivatableObject(trigger);
		end;
		siloTrigger.drawActivate = function(trigger)
			g_currentMission:enableHudIcon("load", 4);
		end;
		siloTrigger.update = function(trigger, dt)
			if trigger.isServer then
				local trailer = trigger.siloTrailer;
				if trigger.activeTriggers >= 4 and trailer ~= nil then
					if trigger.isFilling then
						trailer:resetFillLevelIfNeeded(trigger.selectedFillType);
						local fillLevel = trailer:getFillLevel(trigger.selectedFillType);
						if trailer:allowFillType(trigger.selectedFillType, false) then
							local deltaFillLevel = math.min(trigger.fillLitersPerSecond*0.001*dt, trailer:getFreeCapacity(trigger.selectedFillType), self.fillUnit.level);
							if deltaFillLevel > 0 then
								self:updateFillLevel(-deltaFillLevel);
								trailer:setFillLevel(fillLevel+deltaFillLevel, trigger.selectedFillType, false, trigger.fillVolumeDischargeInfos);
							else
								trigger:setIsFilling(false, FillUtil.FILLTYPE_UNKNOWN);
							end;
						else
							trigger:setIsFilling(false, FillUtil.FILLTYPE_UNKNOWN);
						end;
					end;
				elseif trigger.isFilling then
					trigger:setIsFilling(false, FillUtil.FILLTYPE_UNKNOWN);
				end;
				if trigger.siloTrailerSend ~= trigger.siloTrailer then
					trigger.siloTrailerSend = trigger.siloTrailer;
					trigger:raiseDirtyFlags(trigger.siloTriggerDirtyFlag);
				end;
			end;
		end;
		siloTrigger.startFill = function(trigger, fillType)
			if not trigger.isFilling then
				trigger.activateText = trigger.stopFillText;
				trigger.isFilling = true;
				if trigger.isClient then
					if not trigger.siloFillSoundEnabled and trigger.siloFillSound ~= nil then
						setVisibility(trigger.siloFillSound, true);
						trigger.siloFillSoundEnabled = true;
					end;
					if trigger.scroller then
						setShaderParameter(trigger.scroller, trigger.scrollerShaderParameterName, trigger.scrollerSpeedX, trigger.scrollerSpeedY, 0, 0, false);
					end;
					EffectManager:setFillType(self.pipeEffects, fillType);
					EffectManager:startEffects(self.pipeEffects);
					for i=1, #self.pipeEffects do
						local effect = self.pipeEffects[i];
						local x1,y1,z1 = getWorldTranslation(trigger.siloTrailer.rootNode);
						local x,y,z = getWorldTranslation(effect.node);
						local distance = Utils.vector3Length(x-x1, y-y1, z-z1);
						effect:setUnloadingDistance(distance-1, g_currentMission.terrainRootNode);
					end;
					EffectManager:setFillType(self.conveyorBeltEffects, fillType);
					EffectManager:startEffects(self.conveyorBeltEffects);
					for i=1, #self.conveyorBeltEffects do
						local effect = self.conveyorBeltEffects[i];
						effect:setScrollUpdate(true);
						setVisibility(effect.node, true);
					end;
				end;
			end;
		end;
		siloTrigger.stopFill = function(trigger)
			if trigger.isFilling then
				trigger.isFilling = false;
				trigger.activateText = trigger.startFillText;
				if trigger.isClient then
					if trigger.siloFillSoundEnabled then
						setVisibility(trigger.siloFillSound, false);
						trigger.siloFillSoundEnabled = false;
					end;
					if trigger.scroller then
						setShaderParameter(trigger.scroller, trigger.scrollerShaderParameterName, 0, 0, 0, 0, false);
					end;
					EffectManager:stopEffects(self.pipeEffects);
					for i=1, #self.conveyorBeltEffects do
						local effect = self.conveyorBeltEffects[i];
						effect:setScrollUpdate(false);
						setVisibility(effect.node, false);
					end;
				end;
			end;
		end;
		self.siloTrigger = siloTrigger;
	else
		print("Error: StorageUnit not found tipTrigger or siloTrigger and can't be loaded");
	end;

	return self;
end;

function StorageUnit:finalizePlacement()
	--heap
	local heapNode = getChild(self.rootNode, "heap");
	if heapNode > 0 then
		self.heapNode = heapNode;
		if not MaterialHolder.defaultHeapMaterial then
			MaterialHolder.defaultHeapMaterial = getMaterial(heapNode, 0);
		end;
		local _, heapStartPosY,_ = getTranslation(heapNode);
		local heapMaxPosY = getUserAttribute(heapNode, "moveMaxY");
		self.heapDeltaY = heapMaxPosY - heapStartPosY;
		local shovelFillTriggerIndex = getUserAttribute(heapNode, "shovelFillTriggerIndex");
		if shovelFillTriggerIndex then
			shovelFillTriggerIndex = Utils.indexToObject(heapNode, shovelFillTriggerIndex);
		end;
		if shovelFillTriggerIndex then
			local fillTrigger = ShovelFillTrigger:new();
			if fillTrigger:load(shovelFillTriggerIndex, FillUtil.FILLTYPE_STRAW) then
				g_currentMission:addUpdateable(fillTrigger);
			else
				fillTrigger:delete();
			end;
			fillTrigger.fillShovel = function(fillTrigger, shovel, dt)
				if self.fillUnit.level > 0 then
					local delta = shovel:fillShovelFromTrigger(fillTrigger, self.fillUnit.level, fillTrigger.fillType, dt);
					if delta > 0 then
						self:updateFillLevel(-delta);
					end;
				end;
			end;
			self.shovelFillTrigger = fillTrigger;
		end;
		local shovelTargetIndex = getUserAttribute(heapNode, "shovelTargetIndex");
		if shovelTargetIndex then
			shovelTargetIndex = Utils.indexToObject(heapNode, shovelTargetIndex);
		end;
		if shovelTargetIndex then
			local object = ShovelTarget:new();
			object.nodeId = shovelTargetIndex;
			object.fillTypes = self.parent.acceptedFillTypes;
			g_currentMission:addNodeObject(shovelTargetIndex, object);
			object.addShovelFillLevel = function(object, shovel, fillLevelDelta, fillType)
				if fillLevelDelta > 0 then
					if self.fillUnit.currentFillType ~= FillUtil.FILLTYPE_UNKNOWN and self.fillUnit.currentFillType ~= fillType then
						return 0;
					end;
					if object.fillTypes[fillType] then
						fillLevelDelta = math.min(fillLevelDelta, self.fillUnit.capacity - self.fillUnit.level);
						self:updateFillLevel(fillLevelDelta, fillType);
						return fillLevelDelta;
					end;
				end;
				return 0;
			end;
			self.shovelTarget = object;
		end;
	end;
	--particles
	local xmlFile = loadXMLFile("TempXML", self.parent.configFileName);
	self.pipeEffects = EffectManager:loadEffect(xmlFile, "placeable.pipeEffect", self.rootNode, self);
	self.conveyorBeltEffects = EffectManager:loadEffect(xmlFile, "placeable.conveyorBeltEffect", self.rootNode, self);
	for i=1, #self.conveyorBeltEffects do
		local effect = self.conveyorBeltEffects[i];
		effect.update = function(eff, dt)
			if eff.scrollUpdate then
				eff.scrollPosition = (eff.scrollPosition + dt*eff.scrollSpeed) % eff.scrollLength;
				setShaderParameter(eff.node, "offsetUV", eff.scrollPosition, 0,0,0, false);
			end;
		end;
	end;
	delete(xmlFile);
	--helpers
	local helpersNode = getChild(self.rootNode, "helpers");
	if helpersNode > 0 then
		local helper = Helper:new(helpersNode);
		g_currentMission:addUpdateable(helper);
		helper.text1 = g_i18n:getText("empty");
		helper.text2 = string.format("0 / %d - %d%%", self.fillUnit.capacity, 0);
		self.helper = helper;
	end;
	setShaderParameter(self.siloTrigger.scroller, self.siloTrigger.scrollerShaderParameterName, 0, 0, 0, 0, false);
end;

function StorageUnit:delete()
	if self.tipTrigger then
		self.tipTrigger:unregister(true);
		self.tipTrigger:delete();
	end;
	if self.siloTrigger then
		self.siloTrigger:unregister(true);
		self.siloTrigger:delete();
	end;
	if self.shovelFillTrigger then
		self.shovelFillTrigger:delete();
	end;
	EffectManager:deleteEffects(self.pipeEffects);
	EffectManager:deleteEffects(self.conveyorBeltEffects);
end;

function StorageUnit:setCurrentFillType(fillType)
	self.fillUnit.currentFillType = fillType;
	if fillType ~= FillUtil.FILLTYPE_UNKNOWN then
		if self.tipTrigger then
			self.tipTrigger.acceptedFillTypes = {};
			self.tipTrigger.acceptedFillTypes[fillType] = true;
		end;
		if self.siloTrigger then
			self.siloTrigger.selectedFillType = fillType;
		end;
		if self.shovelFillTrigger then
			self.shovelFillTrigger.fillType = fillType;
		end;
		if self.heapNode then
			local material = MaterialHolder.defaultHeapMaterial;
			local newMaterial = MaterialHolder.heapMaterials[fillType];
			if newMaterial then
				material = newMaterial;
			end;
			setMaterial(self.heapNode, material, 0);
		end;
	else
		if self.tipTrigger then
			self.tipTrigger.acceptedFillTypes = self.parent.acceptedFillTypes;
		end;
		if self.lyingParticleSystems then
			for _, ps in pairs(self.lyingParticleSystems) do
				setVisibility(ps.shape, false);
			end;
		end;
	end;
end;

function StorageUnit:updateFillLevel(delta, fillType)
	if delta ~= 0 then
		if delta > 0 then
			if self.fillUnit.level == 0 or (self.fillUnit.level < 1 and fillType ~= self.fillUnit.currentFillType) then
				self:setCurrentFillType(fillType);
			end;
			self.fillUnit.level = math.min(self.fillUnit.level + delta, self.fillUnit.capacity);
		else
			self.fillUnit.level = math.max(self.fillUnit.level + delta, 0);
			if self.fillUnit.level == 0 then
				self:setCurrentFillType(FillUtil.FILLTYPE_UNKNOWN);
			end;
		end;
		self:heapMoving();
		self.parent.enableUpdateClients = self.index;
		self.parent:raiseDirtyFlags(self.parent.storageDirtyFlag);
	end;
end;

function StorageUnit:heapMoving()
	if self.heapNode then
		local percent = self.fillUnit.level/self.fillUnit.capacity;
		if percent > 0 then
			local x,_,z = getTranslation(self.heapNode);
			setTranslation(self.heapNode, x, self.heapDeltaY*percent, z);
		end;
	end;
	if self.helper then
		local text1 = g_i18n:getText("empty");
		if self.fillUnit.currentFillType ~= FillUtil.FILLTYPE_UNKNOWN then
			text1 = FillUtil.fillTypeIndexToDesc[self.fillUnit.currentFillType].nameI18N;
		end;
		self.helper.text1 = text1;
		local percent = self.fillUnit.level/self.fillUnit.capacity*100;
		self.helper.text2 = string.format("%d / %d - %d%%", self.fillUnit.level, self.fillUnit.capacity, percent);
	end;
end;
